@@ -38,6 +38,7 @@
 #include <unistd.h>
 #include <stdbool.h>
 #include <stdarg.h>
+#include <errno.h>
 #include "igzip_lib.h"		/* Normally you use isa-l.h instead for external programs */
 
 #if defined (HAVE_THREADS)
@@ -224,17 +225,19 @@ size_t get_filesize(FILE * fp)
 	fseeko(fp, 0, SEEK_END);
 #endif
 	fgetpos(fp, &pos);
-	file_size = *(size_t *) & pos;
+	file_size = *(size_t *)&pos;
 	fsetpos(fp, &pos_curr);	/* Restore position */
 
 	return file_size;
 }
 
-uint32_t get_posix_filetime(FILE * fp)
+int get_posix_filetime(FILE * fp, uint32_t * time)
 {
 	struct stat file_stats;
-	fstat(fileno(fp), &file_stats);
-	return file_stats.st_mtime;
+	const int ret = fstat(fileno(fp), &file_stats);
+	if (time != NULL && ret == 0)
+		*time = file_stats.st_mtime;
+	return ret;
 }
 
 uint32_t set_filetime(char *file_name, uint32_t posix_time)
@@ -320,14 +323,15 @@ void *malloc_safe(size_t size)
 	return ptr;
 }
 
-FILE *fopen_safe(char *file_name, char *mode)
+FILE *fopen_safe(const char *file_name, const char *mode)
 {
 	FILE *file;
-	int answer = 0, tmp;
 
 	/* Assumes write mode always starts with w */
 	if (mode[0] == 'w') {
 		if (access(file_name, F_OK) == 0) {
+			int answer = 0, tmp;
+
 			log_print(WARN, "igzip: %s already exists;", file_name);
 			if (is_interactive()) {
 				log_print(WARN, " do you wish to overwrite (y/n)?");
@@ -345,30 +349,14 @@ FILE *fopen_safe(char *file_name, char *mode)
 				log_print(WARN, "       not overwritten\n");
 				return NULL;
 			}
-
-			if (access(file_name, W_OK) != 0) {
-				log_print(ERROR, "igzip: %s: Permission denied\n", file_name);
-				return NULL;
-			}
-		}
-	}
-
-	/* Assumes read mode always starts with r */
-	if (mode[0] == 'r') {
-		if (access(file_name, F_OK) != 0) {
-			log_print(ERROR, "igzip: %s does not exist\n", file_name);
-			return NULL;
-		}
-
-		if (access(file_name, R_OK) != 0) {
-			log_print(ERROR, "igzip: %s: Permission denied\n", file_name);
-			return NULL;
 		}
 	}
 
 	file = fopen(file_name, mode);
 	if (!file) {
-		log_print(ERROR, "igzip: Failed to open %s\n", file_name);
+		const char *error_str = strerror(errno);
+
+		log_print(ERROR, "igzip: Failed to open %s : %s\n", file_name, error_str);
 		return NULL;
 	}
 
@@ -458,17 +446,17 @@ struct thread_pool {
 // Globals for thread pool
 struct thread_pool pool;
 
-static inline int pool_has_space()
+static inline int pool_has_space(void)
 {
 	return ((pool.head + 1) % MAX_JOBQUEUE) != pool.tail;
 }
 
-static inline int pool_has_work()
+static inline int pool_has_work(void)
 {
 	return (pool.queue != pool.head);
 }
 
-int pool_get_work()
+int pool_get_work(void)
 {
 	assert(pool.queue != pool.head);
 	pool.queue = (pool.queue + 1) % MAX_JOBQUEUE;
@@ -544,7 +532,7 @@ void *thread_worker(void *none)
 	pthread_exit(NULL);
 }
 
-int pool_create()
+int pool_create(void)
 {
 	int i;
 	int nthreads = global_options.threads - 1;
@@ -559,7 +547,7 @@ int pool_create()
 	return 0;
 }
 
-void pool_quit()
+void pool_quit(void)
 {
 	int i;
 	pthread_mutex_lock(&pool.mutex);
@@ -583,8 +571,9 @@ int compress_file(void)
 	struct isal_gzip_header gz_hdr;
 	int ret, success = 0;
 
-	char *infile_name = global_options.infile_name, *outfile_name =
-	    global_options.outfile_name;
+	char *infile_name = global_options.infile_name;
+	char *outfile_name = global_options.outfile_name;
+	char *allocated_name = NULL;
 	char *suffix = global_options.suffix;
 	size_t infile_name_len = global_options.infile_name_len;
 	size_t outfile_name_len = global_options.outfile_name_len;
@@ -598,6 +587,7 @@ int compress_file(void)
 	}
 
 	if (infile_name_len == stdin_file_name_len &&
+	    infile_name != NULL &&
 	    memcmp(infile_name, stdin_file_name, infile_name_len) == 0) {
 		infile_name = NULL;
 		infile_name_len = 0;
@@ -605,9 +595,10 @@ int compress_file(void)
 
 	if (outfile_name == NULL && infile_name != NULL && !global_options.use_stdout) {
 		outfile_name_len = infile_name_len + suffix_len;
-		outfile_name = malloc_safe(outfile_name_len + 1);
-		strcpy(outfile_name, infile_name);
-		strcat(outfile_name, suffix);
+		allocated_name = malloc_safe(outfile_name_len + 1);
+		outfile_name = allocated_name;
+		strncpy(outfile_name, infile_name, infile_name_len + 1);
+		strncat(outfile_name, suffix, suffix_len);
 	}
 
 	open_in_file(&in, infile_name);
@@ -615,6 +606,7 @@ int compress_file(void)
 		goto compress_file_cleanup;
 
 	if (infile_name_len != 0 && infile_name_len == outfile_name_len
+	    && infile_name != NULL && outfile_name != NULL
 	    && strncmp(infile_name, outfile_name, infile_name_len) == 0) {
 		log_print(ERROR, "igzip: Error input and output file names must differ\n");
 		goto compress_file_cleanup;
@@ -634,7 +626,8 @@ int compress_file(void)
 
 	isal_gzip_header_init(&gz_hdr);
 	if (global_options.name == NAME_DEFAULT || global_options.name == YES_NAME) {
-		gz_hdr.time = get_posix_filetime(in);
+		if (get_posix_filetime(in, &gz_hdr.time) != 0)
+			goto compress_file_cleanup;
 		gz_hdr.name = infile_name;
 	}
 	gz_hdr.os = UNIX;
@@ -668,7 +661,6 @@ int compress_file(void)
 			size_t outbuf_used = 0;
 			uint8_t *iptr = inbuf;
 			uint8_t *optr = outbuf;
-			ret = 0;
 
 			for (q = 0; q < MAX_JOBQUEUE - 1; q++) {
 				inbuf_used += BLOCK_SIZE;
@@ -741,8 +733,8 @@ int compress_file(void)
 		} while (!end_of_stream);
 
 		// Write gzip trailer
-		ret = fwrite_safe(&crc, sizeof(uint32_t), 1, out, outfile_name);
-		ret += fwrite_safe(&total_in, sizeof(uint32_t), 1, out, outfile_name);
+		fwrite_safe(&crc, sizeof(uint32_t), 1, out, outfile_name);
+		fwrite_safe(&total_in, sizeof(uint32_t), 1, out, outfile_name);
 
 #else // No compiled threading support but asked for threads > 1
 		assert(1);
@@ -788,8 +780,8 @@ int compress_file(void)
 			remove(infile_name);
 	}
 
-	if (global_options.outfile_name == NULL && outfile_name != NULL)
-		free(outfile_name);
+	if (allocated_name != NULL)
+		free(allocated_name);
 
 	return (success == 0);
 }
@@ -804,8 +796,9 @@ int decompress_file(void)
 	const int terminal = 0, implicit = 1, stripped = 2;
 	int ret = 0, success = 0, outfile_type = terminal;
 
-	char *infile_name = global_options.infile_name, *outfile_name =
-	    global_options.outfile_name;
+	char *infile_name = global_options.infile_name;
+	char *outfile_name = global_options.outfile_name;
+	char *allocated_name = NULL;
 	char *suffix = global_options.suffix;
 	size_t infile_name_len = global_options.infile_name_len;
 	size_t outfile_name_len = global_options.outfile_name_len;
@@ -813,7 +806,9 @@ int decompress_file(void)
 	int suffix_index = 0;
 	uint32_t file_time;
 
+	// Allocate mem and setup to hold gzip header info
 	if (infile_name_len == stdin_file_name_len &&
+	    infile_name != NULL &&
 	    memcmp(infile_name, stdin_file_name, infile_name_len) == 0) {
 		infile_name = NULL;
 		infile_name_len = 0;
@@ -849,17 +844,20 @@ int decompress_file(void)
 			outfile_name_len = 0;
 			outfile_type = implicit;
 		}
-		if (outfile_type != terminal)
-			outfile_name = malloc_safe(outfile_name_len >=
-						   MAX_FILEPATH_BUF ? outfile_name_len +
-						   1 : MAX_FILEPATH_BUF);
+		if (outfile_type != terminal) {
+			allocated_name = malloc_safe(outfile_name_len >=
+						     MAX_FILEPATH_BUF ? outfile_name_len +
+						     1 : MAX_FILEPATH_BUF);
+			outfile_name = allocated_name;
+		}
 	}
 
 	open_in_file(&in, infile_name);
 	if (in == NULL)
 		goto decompress_file_cleanup;
 
-	file_time = get_posix_filetime(in);
+	if (get_posix_filetime(in, &file_time) != 0)
+		goto decompress_file_cleanup;
 
 	inbuf_size = global_options.in_buf_size;
 	outbuf_size = global_options.out_buf_size;
@@ -877,6 +875,7 @@ int decompress_file(void)
 	state.next_in = inbuf;
 	state.avail_in = fread_safe(state.next_in, 1, inbuf_size, in, infile_name);
 
+	// Actually read and save the header info
 	ret = isal_read_gzip_header(&state, &gz_hdr);
 	if (ret != ISAL_DECOMP_OK) {
 		log_print(ERROR, "igzip: Error invalid gzip header found for file %s\n",
@@ -887,13 +886,16 @@ int decompress_file(void)
 	if (outfile_type == implicit)
 		file_time = gz_hdr.time;
 
-	if (outfile_type == stripped || (outfile_type == implicit && outfile_name[0] == 0)) {
+	if (outfile_name != NULL && infile_name != NULL
+	    && (outfile_type == stripped
+		|| (outfile_type == implicit && outfile_name[0] == 0))) {
 		outfile_name_len = infile_name_len - suffix_len;
 		memcpy(outfile_name, infile_name, outfile_name_len);
 		outfile_name[outfile_name_len] = 0;
 	}
 
 	if (infile_name_len != 0 && infile_name_len == outfile_name_len
+	    && infile_name != NULL && outfile_name != NULL
 	    && strncmp(infile_name, outfile_name, infile_name_len) == 0) {
 		log_print(ERROR, "igzip: Error input and output file names must differ\n");
 		goto decompress_file_cleanup;
@@ -905,6 +907,7 @@ int decompress_file(void)
 			goto decompress_file_cleanup;
 	}
 
+	// Start reading in compressed data and decompress
 	do {
 		if (state.avail_in == 0) {
 			state.next_in = inbuf;
@@ -926,7 +929,55 @@ int decompress_file(void)
 		if (out != NULL)
 			fwrite_safe(outbuf, 1, state.next_out - outbuf, out, outfile_name);
 
-	} while (!feof(in) || state.avail_out == 0);
+	} while (state.block_state != ISAL_BLOCK_FINISH	// while not done
+		 && (!feof(in) || state.avail_out == 0)	// and work to do
+	    );
+
+	// Add the following to look for and decode additional concatenated files
+	if (!feof(in) && state.avail_in == 0) {
+		state.next_in = inbuf;
+		state.avail_in = fread_safe(state.next_in, 1, inbuf_size, in, infile_name);
+	}
+
+	while (state.avail_in > 0 && state.next_in[0] == 31) {
+		// Look for magic numbers for gzip header. Follows the gzread() decision
+		// whether to treat as trailing junk
+		if (state.avail_in > 1 && state.next_in[1] != 139)
+			break;
+
+		isal_inflate_reset(&state);
+		state.crc_flag = ISAL_GZIP;	// Let isal_inflate() process extra headers
+		do {
+			if (state.avail_in == 0 && !feof(in)) {
+				state.next_in = inbuf;
+				state.avail_in =
+				    fread_safe(state.next_in, 1, inbuf_size, in, infile_name);
+			}
+
+			state.next_out = outbuf;
+			state.avail_out = outbuf_size;
+
+			ret = isal_inflate(&state);
+			if (ret != ISAL_DECOMP_OK) {
+				log_print(ERROR,
+					  "igzip: Error while decompressing extra concatenated"
+					  "gzip files on %s\n", infile_name);
+				goto decompress_file_cleanup;
+			}
+
+			if (out != NULL)
+				fwrite_safe(outbuf, 1, state.next_out - outbuf, out,
+					    outfile_name);
+
+		} while (state.block_state != ISAL_BLOCK_FINISH
+			 && (!feof(in) || state.avail_out == 0));
+
+		if (!feof(in) && state.avail_in == 0) {
+			state.next_in = inbuf;
+			state.avail_in =
+			    fread_safe(state.next_in, 1, inbuf_size, in, infile_name);
+		}
+	}
 
 	if (state.block_state != ISAL_BLOCK_FINISH)
 		log_print(ERROR, "igzip: Error %s does not contain a complete gzip file\n",
@@ -947,8 +998,8 @@ int decompress_file(void)
 			remove(infile_name);
 	}
 
-	if (global_options.outfile_name == NULL && outfile_name != NULL)
-		free(outfile_name);
+	if (allocated_name != NULL)
+		free(allocated_name);
 
 	return (success == 0);
 }
